/**
 * @file TSPSolver.cpp
 * @brief TSP solver (neighborhood search)
 *
 */

#include "TSPSolver.h"
#include <iostream>

bool TSPSolver::solve ( const TSP& tsp , const TSPSolution& initSol , int tabulength , int maxIter , TSPSolution& bestSol )
{
	try
	{
		bool stop = false;
		int  iter = 0;
		
		///Tabu Search
		tabuLength = tabulength;
		tabuList.reserve(tsp.n);
		initTabuList(tsp.n);
		///
		
		TSPSolution currSol(initSol);
		double bestValue, currValue;
		bestValue = currValue = evaluate(currSol,tsp);
		TSPMove move;
		while ( ! stop ) {
		  ++iter;
		  if ( tsp.n < 20 ) currSol.print();
		  std::cout << " (" << iter << ") value " << currValue;// << " (" << evaluate(currSol,tsp) << ")";
		  double aspiration = bestValue-currValue;
		  double bestNeighValue = currValue + findBestNeighbor(tsp,currSol,iter,aspiration,move);
		  std::cout << " move: " << move.from << " , " << move.to;
		  //if ( bestNeighValue < currValue ) {
			//bestValue = currValue = bestNeighValue;
			//currSol = swap(currSol,move);
			//stop = false;
		  //} else {
			//stop = true;
		  //}
		  tabuList[currSol.sequence[move.from]] = iter;
		  tabuList[currSol.sequence[move.to]]   = iter;
		  currValue = bestNeighValue;
		  currSol = swap(currSol,move);
		  if ( currValue < bestValue -0.01 ) {
			bestValue = currValue;
			bestSol = currSol;
			std::cout << " *** "; 
		  }
		  std::cout << std::endl;
		  
		  if ( iter > maxIter || bestNeighValue >= tsp.infinite ) {
			stop = true;
		  }
		  
		}
		//bestSol = currSol;
	}
	catch(std::exception& e)
	{
		std::cout << ">>>EXCEPTION: " << e.what() << std::endl;
		return false;
	}
	return true;
}

TSPSolution& TSPSolver::swap ( TSPSolution& tspSol , const TSPMove& move ) 
{
  TSPSolution tmpSol(tspSol);
  for ( int i = move.from ; i <= move.to ; ++i ) {
	tspSol.sequence[i] = tmpSol.sequence[move.to-(i-move.from)];
  }
  return tspSol;
}


double TSPSolver::findBestNeighbor ( const TSP& tsp , const TSPSolution& currSol , int currIter , double aspiration , TSPMove& move )
{
  double bestDecrement = tsp.infinite;
  for ( uint i = 1 ; i < currSol.sequence.size() - 2 ; i++ ) {
	int a = currSol.sequence[i-1];
	int b = currSol.sequence[i];
//#define L_DEBUG
#ifdef L_DEBUG
	std::cout << "@@@ " << b << ") " << currIter << " - " << tabuList[b] << " <= " << tabuLength << std::endl;
#endif
	for ( uint j = i + 1 ; j < currSol.sequence.size() - 1 ; j++ ) {
	  int e = currSol.sequence[j];
	  int f = currSol.sequence[j+1];
#ifdef L_DEBUG
	  std::cout << "@@@@@@ " << e << ") " << currIter << " - " << tabuList[b] << " <= " << tabuLength << std::endl;
#endif
	  double neighDecrement = - tsp.cost[a][b] - tsp.cost[e][f]
							  + tsp.cost[a][e] + tsp.cost[b][f] ;
#ifdef L_DEBUG
	  {
		std::cout << "*** " << i << " - " << j << "*** a " << a << ", b " << b << ", e " << e << ", f " << f << " -> " << neighDecrement; 
		TSPSolution tmp(currSol); 
		TSPMove m; m.from=i;m.to=j;
		tmp = swap(tmp,m);
		std::cout << " --- "; tmp.print();
		std::cout << " (" << evaluate(tmp,tsp) << ")" << std::endl;
	  }
#endif
	  if ( (currIter - tabuList[b] <= tabuLength) && (currIter - tabuList[e] <= tabuLength) && !(neighDecrement < aspiration-0.01) ) {
		continue;
	  }
	  if ( neighDecrement < bestDecrement ) {
		bestDecrement = neighDecrement;
		move.from = i;
		move.to = j;
	  }
	  //if ( bestDecrement < 0 ) return bestDecrement;
	}
  }
  return bestDecrement;
}
